#ifndef BASIC_FUNCTIONS_H
#define BASIC_FUNCTIONS_H

#include "basic_types.h"
#include <chrono>
#include "pcl_point_type.h"
using namespace chrono;

template <typename T>
Eigen::Matrix<T, 3, 1> RotationQuaternionToEulerVector(
    const Eigen::Quaternion<T>& quaternion) {
  Eigen::Quaternion<T> n_q = quaternion.normalized();
  // We choose the quaternion with positive 'w', i.e., the one with a smaller
  // angle that represents this orientation.
  if (n_q.w() < 0.) {
    // Multiply by -1. http://eigen.tuxfamily.org/bz/show_bug.cgi?id=560
    n_q.w() = -1. * n_q.w();
    n_q.x() = -1. * n_q.x();
    n_q.y() = -1. * n_q.y();
    n_q.z() = -1. * n_q.z();
  }
  // We convert the normalized_quaternion(n_q) into a vector along the rotation
  // axis with length of the rotation angle.
  const T roll = atan2(2 * (n_q.w() * n_q.x() + n_q.y() * n_q.z()),
                       1 - 2 * (n_q.x() * n_q.x() + n_q.y() * n_q.y()));
  const T pitch = asin(2 * (n_q.w() * n_q.y() - n_q.z() * n_q.x()));
  const T yaw = atan2(2 * (n_q.w() * n_q.z() + n_q.x() * n_q.y()),
                      1 - 2 * (n_q.y() * n_q.y() + n_q.z() * n_q.z()));
  return Eigen::Matrix<T, 3, 1>(roll, pitch, yaw);
}


vector<pair<uint32_t, uint32_t>> getIndexPeriod(const uint32_t size,
                                                const uint8_t threadCnt = MAX_THREAD_CNT);

//template<typename ItemType>
//void periodProcessThreadFunc(const PointCloudExport::Ptr &Vector,
//                             pair<uint32_t, uint32_t> period,
//                             boost::function<void(const ItemType &)> processItemFunc)
//{
//    vector<ItemType> tempOutputVec;
//    uint32_t start_index = period.first;
//    uint32_t index_cnt = period.second;
//    uint32_t final_index = start_index + index_cnt;
//    for(uint32_t index = start_index; index < final_index; index++){
//        processItemFunc(Vector->at(index));
//    }
//}

//template<typename ItemType>
//void multiThreadProcess(const PointCloudExport::Ptr &Vector,
//                        boost::function<void(const ItemType &)> processItemFunc)
//{
//    auto start = system_clock::now();
//    vector<pair<uint32_t, uint32_t>> indexPeriod = getIndexPeriod(Vector->size(),
//                                                                  std::thread::hardware_concurrency());
//    vector<ThreadPtr> threadVec;
//    boost::mutex mutex;
//    for(size_t thread_index = 0; thread_index < indexPeriod.size(); thread_index++){
//        threadVec.emplace_back(
//                    ThreadPtr(new boost::thread(boost::bind(&periodProcessThreadFunc<ItemType>,
//                                                            Vector,
//                                                            indexPeriod[thread_index],
//                                                            processItemFunc))));
//    }
//    for(uint8_t thread_index = 0; thread_index < indexPeriod.size(); thread_index++){
//        threadVec[thread_index]->join();
//        threadVec[thread_index].reset();

//    }
//    auto end   = system_clock::now();
//    auto duration = duration_cast<milliseconds>(end - start);
//    LOG(INFO) << "process take " << double(duration.count()) / 1000 << " seconds,"
//              << " output size: " << Vector->size();
//}

template<typename ItemType>
void periodProcessThreadFunc(const vector<ItemType> *Vector,
                             pair<uint32_t, uint32_t> period,
                             boost::function<void(const ItemType &)> processItemFunc)
{
    uint32_t start_index = period.first;
    uint32_t index_cnt = period.second;
    uint32_t final_index = start_index + index_cnt;
    for(uint32_t index = start_index; index < final_index; index++){
        processItemFunc(Vector->at(index));
    }
}

template<typename ItemType>
void multiThreadProcess(const vector<ItemType> &Vector,
                        boost::function<void(const ItemType &)> processItemFunc)
{
    auto start = system_clock::now();
    vector<pair<uint32_t, uint32_t>> indexPeriod = getIndexPeriod(Vector.size(),
                                                                  std::thread::hardware_concurrency());
    vector<ThreadPtr> threadVec;
    boost::mutex mutex;
    for(size_t thread_index = 0; thread_index < indexPeriod.size(); thread_index++){
        threadVec.emplace_back(
                    ThreadPtr(new boost::thread(boost::bind(&periodProcessThreadFunc<ItemType>,
                                                            &Vector,
                                                            indexPeriod[thread_index],
                                                            processItemFunc))));
    }
    for(uint8_t thread_index = 0; thread_index < indexPeriod.size(); thread_index++){
        threadVec[thread_index]->join();
        threadVec[thread_index].reset();

    }
    auto end   = system_clock::now();
    auto duration = duration_cast<milliseconds>(end - start);
    LOG(INFO) << "process take " << double(duration.count()) / 1000 << " seconds,"
              << " output size: " << Vector.size();
}


template<typename ItemType>
void syncProcessThreadFunc(const vector<ItemType> *inputVector,
                           int32_t *currentIndex,
//                           boost::mutex *mutex,
                           boost::function<void(const ItemType &)> processItemFunc)
{
    uint32_t size = inputVector->size();
    int32_t index;
    while(true){
        {
//            boost::unique_lock<boost::mutex> guard(*mutex);
            index = ++(*currentIndex);
        }
        if((uint32_t)index >= size){
//            LOG(INFO) << "index: " << index;
            break;
        }
        processItemFunc(inputVector->at((uint32_t)index));
    }
}

template<typename ItemType>
void syncProcess(const vector<ItemType> &inputVector,
                 boost::function<void(const ItemType &)> processItemFunc)
{
    auto start = system_clock::now();
    int32_t currentIndex = -1;
    boost::mutex mutex;
    vector<ThreadPtr> threadVec;
    for(size_t thread_index = 0; thread_index < std::thread::hardware_concurrency(); thread_index++){
        threadVec.emplace_back(
                    ThreadPtr(new boost::thread(boost::bind(&syncProcessThreadFunc<ItemType>,
                                                            &inputVector,
                                                            &currentIndex,
//                                                            &mutex,
                                                            processItemFunc))));
    }
    for(uint8_t thread_index = 0; thread_index < std::thread::hardware_concurrency(); thread_index++){
        threadVec[thread_index]->join();
        threadVec[thread_index].reset();
    }
    auto end = system_clock::now();
    auto duration = duration_cast<milliseconds>(end - start);
    LOG(INFO) << "process take " << double(duration.count()) / 1000 << " seconds";
}

#endif // BASIC_FUNCTIONS_H
